@@ -1,5 +1,5 @@
 /* chmod -- change permission modes of files
-   Copyright (C) 89, 90, 91, 1995-2003 Free Software Foundation, Inc.
+   Copyright (C) 89, 90, 91, 1995-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,7 +30,6 @@
 #include "modechange.h"
 #include "quote.h"
 #include "root-dev-ino.h"
-#include "savedir.h"
 #include "xfts.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
@@ -40,6 +39,7 @@
 
 enum Change_status
 {
+  CH_NOT_APPLIED,
   CH_SUCCEEDED,
   CH_FAILED,
   CH_NO_CHANGE_REQUESTED
@@ -60,11 +60,11 @@ enum Verbosity
 /* The name the program was run with. */
 char *program_name;
 
-/* If nonzero, change the modes of directories recursively. */
-static int recurse;
+/* If true, change the modes of directories recursively. */
+static bool recurse;
 
-/* If nonzero, force silence (no error messages). */
-static int force_silent;
+/* If true, force silence (no error messages). */
+static bool force_silent;
 
 /* Level of verbosity.  */
 static enum Verbosity verbosity = V_off;
@@ -101,19 +101,30 @@ static struct option const long_options[] =
   {0, 0, 0, 0}
 };
 
-static int
-mode_changed (const char *file, mode_t old_mode)
-{
-  struct stat new_stats;
+/* Return true if the chmodable permission bits of FILE changed.
+   The old mode was OLD_MODE, but it was changed to NEW_MODE.  */
 
-  if (stat (file, &new_stats))
+static bool
+mode_changed (char const *file, mode_t old_mode, mode_t new_mode)
+{
+  if (new_mode & (S_ISUID | S_ISGID | S_ISVTX))
     {
-      if (force_silent == 0)
-	error (0, errno, _("getting new attributes of %s"), quote (file));
-      return 0;
+      /* The new mode contains unusual bits that the call to chmod may
+	 have silently cleared.  Check whether they actually changed.  */
+
+      struct stat new_stats;
+
+      if (stat (file, &new_stats) != 0)
+	{
+	  if (!force_silent)
+	    error (0, errno, _("getting new attributes of %s"), quote (file));
+	  return false;
+	}
+
+      new_mode = new_stats.st_mode;
     }
 
-  return old_mode != new_stats.st_mode;
+  return ((old_mode ^ new_mode) & CHMOD_MODE_BITS) != 0;
 }
 
 /* Tell the user how/if the MODE of FILE has been changed.
@@ -126,6 +137,13 @@ describe_change (const char *file, mode_t mode,
   char perms[11];		/* "-rwxrwxrwx" ls-style modes. */
   const char *fmt;
 
+  if (changed == CH_NOT_APPLIED)
+    {
+      printf (_("neither symbolic link %s nor referent has been changed\n"),
+	      quote (file));
+      return;
+    }
+
   mode_string (mode, perms);
   perms[10] = '\0';		/* `mode_string' does not null terminate. */
   switch (changed)
@@ -143,92 +161,106 @@ describe_change (const char *file, mode_t mode,
       abort ();
     }
   printf (fmt, quote (file),
-	  (unsigned long) (mode & CHMOD_MODE_BITS), &perms[1]);
+	  (unsigned long int) (mode & CHMOD_MODE_BITS), &perms[1]);
 }
 
 /* Change the mode of FILE according to the list of operations CHANGES.
-   Return 0 if successful, 1 if errors occurred.  This function is called
+   Return true if successful.  This function is called
    once for every file system object that fts encounters.  */
 
-static int
+static bool
 process_file (FTS *fts, FTSENT *ent, const struct mode_change *changes)
 {
-  const char *file_full_name = ent->fts_path;
-  const char *file = ent->fts_accpath;
-  const struct stat *sb = ent->fts_statp;
-  mode_t newmode;
-  int errors = 0;
-  int fail;
-  int saved_errno;
+  char const *file_full_name = ent->fts_path;
+  char const *file = ent->fts_accpath;
+  const struct stat *file_stats = ent->fts_statp;
+  mode_t new_mode IF_LINT (= 0);
+  bool ok = true;
+  bool do_chmod;
+  bool symlink_changed = true;
 
   switch (ent->fts_info)
     {
+    case FTS_DP:
+      return true;
+
     case FTS_NS:
       error (0, ent->fts_errno, _("cannot access %s"), quote (file_full_name));
-      return 1;
+      ok = false;
+      break;
 
     case FTS_ERR:
       error (0, ent->fts_errno, _("%s"), quote (file_full_name));
-      return 1;
+      ok = false;
+      break;
 
     case FTS_DNR:
       error (0, ent->fts_errno, _("cannot read directory %s"),
 	     quote (file_full_name));
-      return 1;
+      ok = false;
+      break;
 
     default:
       break;
     }
 
-  /* If this is the second (post-order) encounter with a directory,
-     then return right away.  */
-  if (ent->fts_info == FTS_DP)
-    return 0;
+  do_chmod = ok;
 
-  if (ROOT_DEV_INO_CHECK (root_dev_ino, sb))
+  if (do_chmod && ROOT_DEV_INO_CHECK (root_dev_ino, file_stats))
     {
       ROOT_DEV_INO_WARN (file_full_name);
-      return 1;
+      ok = false;
+      do_chmod = false;
     }
 
-  if (S_ISLNK (sb->st_mode))
-    return 0;
-
-  newmode = mode_adjust (sb->st_mode, changes);
-
-  fail = chmod (file, newmode);
-  saved_errno = errno;
-
-  if (verbosity == V_high
-      || (verbosity == V_changes_only
-	  && !fail && mode_changed (file, sb->st_mode)))
-    describe_change (file_full_name, newmode,
-		     (fail ? CH_FAILED : CH_SUCCEEDED));
-
-  if (fail)
+  if (do_chmod)
     {
-      if (force_silent == 0)
-	error (0, saved_errno, _("changing permissions of %s"),
-	       quote (file_full_name));
-      errors = 1;
+      new_mode = mode_adjust (file_stats->st_mode, changes);
+
+      if (S_ISLNK (file_stats->st_mode))
+	symlink_changed = false;
+      else
+	{
+	  ok = (chmod (file, new_mode) == 0);
+
+	  if (! (ok | force_silent))
+	    error (0, errno, _("changing permissions of %s"),
+		   quote (file_full_name));
+	}
+    }
+
+  if (verbosity != V_off)
+    {
+      bool changed =
+	(ok && symlink_changed
+	 && mode_changed (file, file_stats->st_mode, new_mode));
+
+      if (changed || verbosity == V_high)
+	{
+	  enum Change_status ch_status =
+	    (!ok ? CH_FAILED
+	     : !symlink_changed ? CH_NOT_APPLIED
+	     : !changed ? CH_NO_CHANGE_REQUESTED
+	     : CH_SUCCEEDED);
+	  describe_change (file_full_name, new_mode, ch_status);
+	}
     }
 
   if ( ! recurse)
     fts_set (fts, ent, FTS_SKIP);
 
-  return errors;
+  return ok;
 }
 
 /* Recursively change the modes of the specified FILES (the last entry
    of which is NULL) according to the list of operations CHANGES.
    BIT_FLAGS controls how fts works.
-   If the fts_open call fails, exit nonzero.
-   Otherwise, return nonzero upon error.  */
+   Return true if successful.  */
 
-static int
+static bool
 process_files (char **files, int bit_flags, const struct mode_change *changes)
 {
-  int fail = 0;
+  bool ok = true;
 
   FTS *fts = xfts_open (files, bit_flags, NULL);
 
@@ -243,12 +275,12 @@ process_files (char **files, int bit_flags, const struct mode_change *changes)
 	    {
 	      /* FIXME: try to give a better message  */
 	      error (0, errno, _("fts_read failed"));
-	      fail = 1;
+	      ok = false;
 	    }
 	  break;
 	}
 
-      fail |= process_file (fts, ent, changes);
+      ok &= process_file (fts, ent, changes);
     }
 
   /* Ignore failure, since the only way it can do so is in failing to
@@ -256,13 +288,13 @@ process_files (char **files, int bit_flags, const struct mode_change *changes)
      that doesn't matter.  */
   fts_close (fts);
 
-  return fail;
+  return ok;
 }
 
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -307,9 +339,10 @@ int
 main (int argc, char **argv)
 {
   struct mode_change *changes;
-  int fail = 0;
-  int modeind = 0;		/* Index of the mode argument in `argv'. */
-  int thisind;
+  char *mode = NULL;
+  size_t mode_len = 0;
+  size_t mode_alloc = 0;
+  bool ok;
   bool preserve_root = false;
   int c;
 
@@ -321,20 +354,15 @@ main (int argc, char **argv)
 
   atexit (close_stdout);
 
-  recurse = force_silent = 0;
+  recurse = force_silent = false;
 
-  while (1)
+  while ((c = getopt_long (argc, argv,
+			   "Rcfvr::w::x::X::s::t::u::g::o::a::,::+::-::=::",
+			   long_options, NULL))
+	 != -1)
     {
-      thisind = optind ? optind : 1;
-
-      c = getopt_long (argc, argv, "RcfvrwxXstugoa,+-=", long_options, NULL);
-      if (c == -1)
-	break;
-
       switch (c)
 	{
-	case 0:
-	  break;
 	case 'r':
 	case 'w':
 	case 'x':
@@ -349,15 +377,25 @@ main (int argc, char **argv)
 	case '+':
 	case '-':
 	case '=':
-	  if (modeind != 0 && modeind != thisind)
-	    {
-	      static char char_string[2] = {0, 0};
-	      char_string[0] = c;
-	      error (EXIT_FAILURE, 0,
-		     _("invalid character %s in mode string %s"),
-		     quote_n (0, char_string), quote_n (1, argv[thisind]));
-	    }
-	  modeind = thisind;
+	  {
+	    /* Allocate a mode string (e.g., "-rwx") by concatenating
+	       the argument containing this option.  If a previous mode
+	       string was given, concatenate the previous string, a
+	       comma, and the new string (e.g., "-s,-rwx").  */
+
+	    char const *arg = argv[optind - 1];
+	    size_t arg_len = strlen (arg);
+	    size_t mode_comma_len = mode_len + !!mode_len;
+	    size_t new_mode_len = mode_comma_len + arg_len;
+	    if (mode_alloc <= new_mode_len)
+	      {
+		mode_alloc = new_mode_len + 1;
+		mode = x2realloc (mode, &mode_alloc);
+	      }
+	    mode[mode_len] = ',';
+	    strcpy (mode + mode_comma_len, arg);
+	    mode_len = new_mode_len;
+	  }
 	  break;
 	case NO_PRESERVE_ROOT:
 	  preserve_root = false;
@@ -369,13 +407,13 @@ main (int argc, char **argv)
 	  reference_file = optarg;
 	  break;
 	case 'R':
-	  recurse = 1;
+	  recurse = true;
 	  break;
 	case 'c':
 	  verbosity = V_changes_only;
 	  break;
 	case 'f':
-	  force_silent = 1;
+	  force_silent = true;
 	  break;
 	case 'v':
 	  verbosity = V_high;
@@ -387,28 +425,39 @@ main (int argc, char **argv)
 	}
     }
 
-  if (modeind == 0 && reference_file == NULL)
-    modeind = optind++;
+  if (reference_file)
+    {
+      if (mode)
+	error (EXIT_FAILURE, 0,
+	       _("cannot combine mode and --reference options"));
+    }
+  else
+    {
+      if (!mode)
+	mode = argv[optind++];
+    }
 
   if (optind >= argc)
     {
-      error (0, 0, _("too few arguments"));
+      if (!mode || mode != argv[optind - 1])
+	error (0, 0, _("missing operand"));
+      else
+	error (0, 0, _("missing operand after %s"), quote (argv[argc - 1]));
       usage (EXIT_FAILURE);
     }
 
   changes = (reference_file ? mode_create_from_ref (reference_file)
-	     : mode_compile (argv[modeind], MODE_MASK_ALL));
+	     : mode_compile (mode, MODE_MASK_ALL));
 
   if (changes == MODE_INVALID)
-    error (EXIT_FAILURE, 0,
-	   _("invalid mode string: %s"), quote (argv[modeind]));
+    error (EXIT_FAILURE, 0, _("invalid mode: %s"), quote (mode));
   else if (changes == MODE_MEMORY_EXHAUSTED)
     xalloc_die ();
   else if (changes == MODE_BAD_REFERENCE)
     error (EXIT_FAILURE, errno, _("failed to get attributes of %s"),
 	   quote (reference_file));
 
-  if (recurse && preserve_root)
+  if (recurse & preserve_root)
     {
       static struct dev_ino dev_ino_buf;
       root_dev_ino = get_root_dev_ino (&dev_ino_buf);
@@ -421,7 +470,7 @@ main (int argc, char **argv)
       root_dev_ino = NULL;
     }
 
-  fail = process_files (argv + optind, FTS_COMFOLLOW, changes);
+  ok = process_files (argv + optind, FTS_COMFOLLOW, changes);
 
-  exit (fail ? EXIT_FAILURE : EXIT_SUCCESS);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
