@@ -83,7 +83,7 @@ src_to_dest_free (void *x)
 
 /* Remove the entry matching INO/DEV from the table
    that maps source ino/dev to destination file name.  */
-void
+extern void
 forget_created (ino_t ino, dev_t dev)
 {
   struct Src_to_dest probe;
@@ -99,9 +99,9 @@ forget_created (ino_t ino, dev_t dev)
 }
 
 /* Add PATH to the list of files that we have created.
-   Return 1 if we can't stat PATH, otherwise 0.  */
+   Return true if successful.  */
 
-int
+extern bool
 remember_created (const char *path)
 {
   struct stat sb;
@@ -109,17 +109,17 @@ remember_created (const char *path)
   if (stat (path, &sb) < 0)
     {
       error (0, errno, "%s", quote (path));
-      return 1;
+      return false;
     }
 
   remember_copied (path, sb.st_ino, sb.st_dev);
-  return 0;
+  return true;
 }
 
 /* If INO/DEV correspond to an already-copied source file, return the
    name of the corresponding destination file.  Otherwise, return NULL.  */
 
-char *
+extern char *
 src_to_dest_lookup (ino_t ino, dev_t dev)
 {
   struct Src_to_dest ent;
@@ -134,7 +134,7 @@ src_to_dest_lookup (ino_t ino, dev_t dev)
    to the list of files we have copied.
    Return NULL if inserted, otherwise non-NULL. */
 
-char *
+extern char *
 remember_copied (const char *name, ino_t ino, dev_t dev)
 {
   struct Src_to_dest *ent;
@@ -166,7 +166,7 @@ remember_copied (const char *name, ino_t ino, dev_t dev)
 }
 
 /* Initialize the hash table.  */
-void
+extern void
 hash_init (void)
 {
   src_to_dest = hash_initialize (INITIAL_TABLE_SIZE, NULL,
@@ -180,7 +180,7 @@ hash_init (void)
 /* Reset the hash structure in the global variable `htab' to
    contain no entries.  */
 
-void
+extern void
 forget_all (void)
 {
   hash_free (src_to_dest);
