@@ -1,5 +1,5 @@
 /* remove.c -- core functions for removing files and directories
-   Copyright (C) 88, 90, 91, 1994-2003 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1994-2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -35,6 +35,8 @@
 #include "obstack.h"
 #include "quote.h"
 #include "remove.h"
+#include "root-dev-ino.h"
+#include "yesno.h"
 
 /* Avoid shadowing warnings because these are functions declared
    in dirname.h as well as locals used below.  */
@@ -63,7 +65,7 @@ enum Ternary
   };
 typedef enum Ternary Ternary;
 
-/* The prompt function may be called twice a given directory.
+/* The prompt function may be called twice for a given directory.
    The first time, we ask whether to descend into it, and the
    second time, we ask whether to remove it.  */
 enum Prompt_action
@@ -79,14 +81,6 @@ int rpl_lstat (const char *, struct stat *);
 # define lstat(Name, Stat_buf) rpl_lstat(Name, Stat_buf)
 #endif
 
-#ifdef D_INO_IN_DIRENT
-# define D_INO(dp) ((dp)->d_ino)
-# define ENABLE_CYCLE_CHECK
-#else
-/* Some systems don't have inodes, so fake them to avoid lots of ifdefs.  */
-# define D_INO(dp) 1
-#endif
-
 /* Initial capacity of per-directory hash table of entries that have
    been processed but not been deleted.  */
 #define HT_UNREMOVABLE_INITIAL_CAPACITY 13
@@ -106,21 +100,12 @@ struct AD_ent
      because the user declined.  */
   enum RM_status status;
 
-  union
-  {
-    /* The directory's dev/ino.  Used to ensure that `chdir some-subdir', then
-       `chdir ..' takes us back to the same directory from which we started).
-       (valid for all but the bottommost entry on the stack.  */
-    struct dev_ino a;
-
-    /* Enough information to restore the initial working directory.
-       (valid only for the bottommost entry on the stack)  */
-    struct saved_cwd saved_cwd;
-  } u;
+  /* The directory's dev/ino.  Used to ensure that `chdir some-subdir', then
+     `chdir ..' takes us back to the same directory from which we started).
+     (valid for all but the bottommost entry on the stack.  */
+  struct dev_ino dev_ino;
 };
 
-int yesno ();
-
 extern char *program_name;
 
 struct dirstack_state
@@ -148,27 +133,42 @@ struct dirstack_state
   /* Used to detect cycles.  */
   struct cycle_check_state cycle_check_state;
 
-  /* Target of a longjmp in case rm detects a directory cycle.  */
+  /* Target of a longjmp in case rm has to stop processing the current
+     command-line argument.  This happens 1) when rm detects a directory
+     cycle or 2) when it has processed one or more directories, but then
+     is unable to return to the initial working directory to process
+     additional `.'-relative command-line arguments.  */
   jmp_buf current_arg_jumpbuf;
 };
 typedef struct dirstack_state Dirstack_state;
 
-Dirstack_state *
+struct cwd_state
+{
+  /* The value of errno after a failed save_cwd or restore_cwd.  */
+  int saved_errno;
+
+  /* Information (open file descriptor or absolute directory name)
+     required in order to restore the initial working directory.  */
+  struct saved_cwd saved_cwd;
+};
+
+static Dirstack_state *
 ds_init ()
 {
-  Dirstack_state *ds = XMALLOC (struct dirstack_state, 1);
+  Dirstack_state *ds = xmalloc (sizeof *ds);
   obstack_init (&ds->dir_stack);
   obstack_init (&ds->len_stack);
   obstack_init (&ds->Active_dir);
   return ds;
 }
 
-void
+static void
 ds_free (Dirstack_state *ds)
 {
   obstack_free (&ds->dir_stack, NULL);
   obstack_free (&ds->len_stack, NULL);
   obstack_free (&ds->Active_dir, NULL);
+  free (ds);
 }
 
 static void
@@ -186,9 +186,7 @@ hash_compare_strings (void const *x, void const *y)
 static inline void
 push_dir (Dirstack_state *ds, const char *dir_name)
 {
-  size_t len;
-
-  len = strlen (dir_name);
+  size_t len = strlen (dir_name);
 
   /* Append the string onto the stack.  */
   obstack_grow (&ds->dir_stack, dir_name, len);
@@ -208,8 +206,8 @@ push_dir (Dirstack_state *ds, const char *dir_name)
 static inline char *
 top_dir (Dirstack_state const *ds)
 {
-  int n_lengths = obstack_object_size (&ds->len_stack) / sizeof (size_t);
-  size_t *length = (size_t *) obstack_base (&ds->len_stack);
+  size_t n_lengths = obstack_object_size (&ds->len_stack) / sizeof (size_t);
+  size_t *length = obstack_base (&ds->len_stack);
   size_t top_len = length[n_lengths - 1];
   char const *p = obstack_next_free (&ds->dir_stack) - top_len;
   char *q = xmalloc (top_len);
@@ -221,8 +219,8 @@ top_dir (Dirstack_state const *ds)
 static inline void
 pop_dir (Dirstack_state *ds)
 {
-  int n_lengths = obstack_object_size (&ds->len_stack) / sizeof (size_t);
-  size_t *length = (size_t *) obstack_base (&ds->len_stack);
+  size_t n_lengths = obstack_object_size (&ds->len_stack) / sizeof (size_t);
+  size_t *length = obstack_base (&ds->len_stack);
   size_t top_len;
 
   assert (n_lengths > 0);
@@ -240,13 +238,13 @@ pop_dir (Dirstack_state *ds)
 
 /* Copy the SRC_LEN bytes of data beginning at SRC into the DST_LEN-byte
    buffer, DST, so that the last source byte is at the end of the destination
-   buffer.  If SRC_LEN is longer than DST_LEN, then set *TRUNCATED to non-zero.
+   buffer.  If SRC_LEN is longer than DST_LEN, then set *TRUNCATED.
    Set *RESULT to point to the beginning of (the portion of) the source data
    in DST.  Return the number of bytes remaining in the destination buffer.  */
 
 static size_t
 right_justify (char *dst, size_t dst_len, const char *src, size_t src_len,
-	       char **result, int *truncated)
+	       char **result, bool *truncated)
 {
   const char *sp;
   char *dp;
@@ -255,14 +253,14 @@ right_justify (char *dst, size_t dst_len, const char *src, size_t src_len,
     {
       sp = src;
       dp = dst + (dst_len - src_len);
-      *truncated = 0;
+      *truncated = false;
     }
   else
     {
       sp = src + (src_len - dst_len);
       dp = dst;
       src_len = dst_len;
-      *truncated = 1;
+      *truncated = true;
     }
 
   *result = memcpy (dp, sp, src_len);
@@ -281,8 +279,8 @@ full_filename_ (Dirstack_state const *ds, const char *filename)
   static char *buf = NULL;
   static size_t n_allocated = 0;
 
-  int dir_len = obstack_object_size (&ds->dir_stack);
-  char *dir_name = (char *) obstack_base (&ds->dir_stack);
+  size_t dir_len = obstack_object_size (&ds->dir_stack);
+  char *dir_name = obstack_base (&ds->dir_stack);
   size_t n_bytes_needed;
   size_t filename_len;
 
@@ -296,18 +294,19 @@ full_filename_ (Dirstack_state const *ds, const char *filename)
 	 error involving a file name to be expanded here wouldn't ever
 	 be issued.  Use realloc and fall back on using a static buffer
 	 if memory allocation fails.  */
-      buf = realloc (buf, n_bytes_needed);
+      char *new_buf = realloc (buf, n_bytes_needed);
       n_allocated = n_bytes_needed;
 
-      if (buf == NULL)
+      if (new_buf == NULL)
 	{
 #define SBUF_SIZE 512
 #define ELLIPSES_PREFIX "[...]"
 	  static char static_buf[SBUF_SIZE];
-	  int truncated;
+	  bool truncated;
 	  size_t len;
 	  char *p;
 
+	  free (buf);
 	  len = right_justify (static_buf, SBUF_SIZE, filename,
 			       filename_len + 1, &p, &truncated);
 	  right_justify (static_buf, len, dir_name, dir_len, &p, &truncated);
@@ -318,6 +317,8 @@ full_filename_ (Dirstack_state const *ds, const char *filename)
 	    }
 	  return p;
 	}
+
+      buf = new_buf;
     }
 
   if (filename_len == 1 && *filename == '.' && dir_len)
@@ -366,55 +367,64 @@ AD_stack_pop (Dirstack_state *ds)
 /* chdir `up' one level.
    Whenever using chdir '..', verify that the post-chdir
    dev/ino numbers for `.' match the saved ones.
-   Return the name (in malloc'd storage) of the
-   directory (usually now empty) from which we're coming.  */
-static char *
-AD_pop_and_chdir (Dirstack_state *ds)
+   If they don't match, exit nonzero.
+   Set *PREV_DIR to the name (in malloc'd storage) of the
+   directory (usually now empty) from which we're coming.
+   If we're at the bottom of the AD stack (about to return
+   to the initial working directory), then use CWD.
+   If that restore_cwd fails, set CWD_STATE->saved_errno.  */
+static void
+AD_pop_and_chdir (Dirstack_state *ds, char **prev_dir,
+		  struct cwd_state *cwd_state)
 {
-  /* Get the name of the current directory from the top of the stack.  */
-  char *dir = top_dir (ds);
   enum RM_status old_status = AD_stack_top(ds)->status;
-  struct stat sb;
   struct AD_ent *top;
 
+  /* Get the name of the current (but soon to be `previous') directory
+     from the top of the stack.  */
+  *prev_dir = top_dir (ds);
+
   AD_stack_pop (ds);
+  top = AD_stack_top (ds);
 
   /* Propagate any failure to parent.  */
-  UPDATE_STATUS (AD_stack_top(ds)->status, old_status);
+  UPDATE_STATUS (top->status, old_status);
 
   assert (AD_stack_height (ds));
 
-  top = AD_stack_top (ds);
   if (1 < AD_stack_height (ds))
     {
+      struct stat sb;
+
       /* We can give a better diagnostic here, since the target is relative. */
-      if (chdir (".."))
+      if (chdir ("..") != 0)
 	{
 	  error (EXIT_FAILURE, errno,
 		 _("cannot chdir from %s to .."),
 		 quote (full_filename (".")));
 	}
-    }
-  else
-    {
-      if (restore_cwd (&top->u.saved_cwd))
+
+      if (lstat (".", &sb))
 	error (EXIT_FAILURE, errno,
-	       _("failed to return to initial working directory"));
-    }
+	       _("cannot lstat `.' in %s"), quote (full_filename (".")));
 
-  if (lstat (".", &sb))
-    error (EXIT_FAILURE, errno,
-	   _("cannot lstat `.' in %s"), quote (full_filename (".")));
-
-  if (1 < AD_stack_height (ds))
-    {
       /*  Ensure that post-chdir dev/ino match the stored ones.  */
-      if ( ! SAME_INODE (sb, top->u.a))
+      if ( ! SAME_INODE (sb, top->dev_ino))
 	error (EXIT_FAILURE, 0,
 	       _("%s changed dev/ino"), quote (full_filename (".")));
     }
-
-  return dir;
+  else
+    {
+      if (restore_cwd (&cwd_state->saved_cwd) != 0)
+	{
+	  /* We've failed to return to the initial working directory.
+	     That failure may be harmless if x->require_restore_cwd is false,
+	     but we do have to remember that fact, including the errno value,
+	     so we can give an accurate diagnostic when reporting the failure
+	     to remove a subsequent relative-named command-line argument.  */
+	  cwd_state->saved_errno = errno;
+	}
+    }
 }
 
 /* Initialize *HT if it is NULL.
@@ -423,10 +433,12 @@ static void
 AD_mark_helper (Hash_table **ht, char const *filename)
 {
   if (*ht == NULL)
-    *ht = hash_initialize (HT_UNREMOVABLE_INITIAL_CAPACITY, NULL, hash_pjw,
-			   hash_compare_strings, hash_freer);
-  if (*ht == NULL)
-    xalloc_die ();
+    {
+      *ht = hash_initialize (HT_UNREMOVABLE_INITIAL_CAPACITY, NULL, hash_pjw,
+			     hash_compare_strings, hash_freer);
+      if (*ht == NULL)
+	xalloc_die ();
+    }
   if (! hash_insert (*ht, filename))
     xalloc_die ();
 }
@@ -454,10 +466,10 @@ AD_mark_current_as_unremovable (Dirstack_state *ds)
   AD_mark_helper (&top->unremovable, curr);
 }
 
-/* Push the initial cwd info onto the stack.
+/* Push an initial dummy entry onto the stack.
    This will always be the bottommost entry on the stack.  */
 static void
-AD_push_initial (Dirstack_state *ds, struct saved_cwd const *cwd)
+AD_push_initial (Dirstack_state *ds)
 {
   struct AD_ent *top;
 
@@ -466,8 +478,13 @@ AD_push_initial (Dirstack_state *ds, struct saved_cwd const *cwd)
 
   /* Fill in the new values.  */
   top = AD_stack_top (ds);
-  top->u.saved_cwd = *cwd;
   top->unremovable = NULL;
+
+  /* These should never be used.
+     Give them values that might look suspicious
+     in a debugger or in a diagnostic.  */
+  top->dev_ino.st_dev = TYPE_MAXIMUM (dev_t);
+  top->dev_ino.st_ino = TYPE_MAXIMUM (ino_t);
 }
 
 /* Push info about the current working directory (".") onto the
@@ -488,7 +505,7 @@ AD_push (Dirstack_state *ds, char const *dir,
 	   _("cannot lstat `.' in %s"), quote (full_filename (".")));
 
   if ( ! SAME_INODE (sb, *dir_sb_from_parent))
-    error (EXIT_FAILURE, errno,
+    error (EXIT_FAILURE, 0,
 	   _("%s changed dev/ino"), quote (full_filename (".")));
 
   /* Extend the stack.  */
@@ -496,44 +513,32 @@ AD_push (Dirstack_state *ds, char const *dir,
 
   /* Fill in the new values.  */
   top = AD_stack_top (ds);
-  top->u.a.st_dev = sb.st_dev;
-  top->u.a.st_ino = sb.st_ino;
+  top->dev_ino.st_dev = sb.st_dev;
+  top->dev_ino.st_ino = sb.st_ino;
   top->unremovable = NULL;
 }
 
-static int
+static bool
 AD_is_removable (Dirstack_state const *ds, char const *file)
 {
   struct AD_ent *top = AD_stack_top (ds);
   return ! (top->unremovable && hash_lookup (top->unremovable, file));
 }
 
-/* A wrapper for readdir so that callers don't see entries for `.' or `..'.  */
-static struct dirent *
-readdir_ignoring_dotdirs (DIR *dirp)
-{
-  while (1)
-    {
-      struct dirent *dp = readdir (dirp);
-      if (dp == NULL || ! DOT_OR_DOTDOT (dp->d_name))
-	return dp;
-    }
-}
-
-/* Return nonzero if DIR is determined to be an empty directory
+/* Return true if DIR is determined to be an empty directory
    or if opendir or readdir fails.  */
 static bool
 is_empty_dir (char const *dir)
 {
   DIR *dirp = opendir (dir);
-  struct dirent *dp;
+  struct dirent const *dp;
   int saved_errno;
 
   if (dirp == NULL)
     return false;
 
   errno = 0;
-  dp = readdir_ignoring_dotdirs (dirp);
+  dp = readdir_ignoring_dot_and_dotdot (dirp);
   saved_errno = errno;
   closedir (dirp);
   if (dp != NULL)
@@ -541,6 +546,23 @@ is_empty_dir (char const *dir)
   return saved_errno == 0 ? true : false;
 }
 
+/* Return true if FILE is not a symbolic link and it is not writable.
+   Also return true if FILE cannot be lstat'ed.  Otherwise, return false.
+   If lstat succeeds, set *BUF_P to BUF.
+   This is to avoid calling euidaccess when FILE is a symlink.  */
+static bool
+write_protected_non_symlink (char const *file,
+			     struct stat **buf_p,
+			     struct stat *buf)
+{
+  if (lstat (file, buf) != 0)
+    return false;
+  *buf_p = buf;
+  if (S_ISLNK (buf->st_mode))
+    return false;
+  return euidaccess (file, W_OK) == -1 && errno == EACCES;
+}
+
 /* Prompt whether to remove FILENAME, if required via a combination of
    the options specified by X and/or file attributes.  If the file may
    be removed, return RM_OK.  If the user declines to remove the file,
@@ -559,41 +581,55 @@ prompt (Dirstack_state const *ds, char const *filename,
 	struct rm_options const *x, enum Prompt_action mode,
 	Ternary *is_dir, Ternary *is_empty)
 {
-  int write_protected = 0;
+  bool write_protected = false;
+  struct stat *sbuf = NULL;
+  struct stat buf;
+
   *is_empty = T_UNKNOWN;
   *is_dir = T_UNKNOWN;
 
-  if ((!x->ignore_missing_files && (x->interactive || x->stdin_tty)
-       && (write_protected = (euidaccess (filename, W_OK) && errno == EACCES)))
+  if (((!x->ignore_missing_files & (x->interactive | x->stdin_tty))
+       && (write_protected = write_protected_non_symlink (filename,
+							  &sbuf, &buf)))
       || x->interactive)
     {
-      struct stat sbuf;
-      if (lstat (filename, &sbuf))
+      if (sbuf == NULL)
 	{
-	  /* lstat failed.  This happens e.g., with `rm '''.  */
-	  error (0, errno, _("cannot lstat %s"),
+	  sbuf = &buf;
+	  if (lstat (filename, sbuf))
+	    {
+	      /* lstat failed.  This happens e.g., with `rm '''.  */
+	      error (0, errno, _("cannot lstat %s"),
+		     quote (full_filename (filename)));
+	      return RM_ERROR;
+	    }
+	}
+
+      if (S_ISDIR (sbuf->st_mode) && !x->recursive)
+	{
+	  error (0, EISDIR, _("cannot remove directory %s"),
 		 quote (full_filename (filename)));
 	  return RM_ERROR;
 	}
 
       /* Using permissions doesn't make sense for symlinks.  */
-      if (S_ISLNK (sbuf.st_mode))
+      if (S_ISLNK (sbuf->st_mode))
 	{
 	  if ( ! x->interactive)
 	    return RM_OK;
-	  write_protected = 0;
+	  write_protected = false;
 	}
 
       /* Issue the prompt.  */
       {
 	char const *quoted_name = quote (full_filename (filename));
 
-	*is_dir = (S_ISDIR (sbuf.st_mode) ? T_YES : T_NO);
+	*is_dir = (S_ISDIR (sbuf->st_mode) ? T_YES : T_NO);
 
 	/* FIXME: use a variant of error (instead of fprintf) that doesn't
 	   append a newline.  Then we won't have to declare program_name in
 	   this file.  */
-	if (S_ISDIR (sbuf.st_mode)
+	if (S_ISDIR (sbuf->st_mode)
 	    && x->recursive
 	    && mode == PA_DESCEND_INTO_DIR
 	    && ((*is_empty = (is_empty_dir (filename) ? T_YES : T_NO))
@@ -613,7 +649,7 @@ prompt (Dirstack_state const *ds, char const *filename,
 		     (write_protected
 		      ? _("%s: remove write-protected %s %s? ")
 		      : _("%s: remove %s %s? ")),
-		     program_name, file_type (&sbuf), quoted_name);
+		     program_name, file_type (sbuf), quoted_name);
 	  }
 
 	if (!yesno ())
@@ -770,14 +806,12 @@ remove_entry (Dirstack_state const *ds, char const *filename,
 
   DO_UNLINK (filename, x);
 
-  /* Accept either EISDIR or EPERM as an indication that FILENAME may be
-     a directory.  POSIX says that unlink must set errno to EPERM when it
-     fails to remove a directory, while Linux-2.4.18 sets it to EISDIR.  */
-  if ((errno != EISDIR && errno != EPERM) || ! x->recursive)
+  if (! x->recursive
+      || errno == ENOENT || errno == ENOTDIR
+      || errno == ELOOP || errno == ENAMETOOLONG)
     {
-      /* some other error code.  Report it and fail.
-	 Likewise, if we're trying to remove a directory without
-	 the --recursive option.  */
+      /* Either --recursive is not in effect, or the file cannot be a
+	 directory.  Report the unlink problem and fail.  */
       error (0, errno, _("cannot remove %s"),
 	     quote (full_filename (filename)));
       return RM_ERROR;
@@ -821,14 +855,14 @@ remove_cwd_entries (Dirstack_state *ds, char **subdir, struct stat *subdir_sb,
 
   while (1)
     {
-      struct dirent *dp;
+      struct dirent const *dp;
       enum RM_status tmp_status;
       const char *f;
 
       /* Set errno to zero so we can distinguish between a readdir failure
 	 and when readdir simply finds that there are no more entries.  */
       errno = 0;
-      if ((dp = readdir_ignoring_dotdirs (dirp)) == NULL)
+      if ((dp = readdir_ignoring_dot_and_dotdot (dirp)) == NULL)
 	{
 	  if (errno)
 	    {
@@ -879,8 +913,7 @@ remove_cwd_entries (Dirstack_state *ds, char **subdir, struct stat *subdir_sb,
 	case RM_NONEMPTY_DIR:
 	  {
 	    /* Save a copy of errno, in case the preceding unlink (from
-	       remove_entry's DO_UNLINK) of a non-directory failed due
-	       to EPERM.  */
+	       remove_entry's DO_UNLINK) of a non-directory failed.  */
 	    int saved_errno = errno;
 
 	    /* Record dev/ino of F so that we can compare
@@ -890,7 +923,8 @@ remove_cwd_entries (Dirstack_state *ds, char **subdir, struct stat *subdir_sb,
 	      error (EXIT_FAILURE, errno, _("cannot lstat %s"),
 		     quote (full_filename (f)));
 
-	    if (chdir (f))
+	    errno = ENOTDIR;
+	    if (! S_ISDIR (subdir_sb->st_mode) || chdir (f) != 0)
 	      {
 		/* It is much more common that we reach this point for an
 		   inaccessible directory.  Hence the second diagnostic, below.
@@ -959,10 +993,11 @@ The following directory is part of the cycle:\n  %s\n"),
     return to the same directory from which we came, if necessary.
     Return 1 for success, 0 if some file cannot be removed or if
     a chdir fails.
-    If the working directory cannot be restored, exit immediately.  */
+    If the initial working directory cannot be saved or restored,
+    record the offending errno value in (*CWD_STATE)->saved_errno.  */
 
 static enum RM_status
-remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
+remove_dir (Dirstack_state *ds, char const *dir, struct cwd_state **cwd_state,
 	    struct rm_options const *x)
 {
   enum RM_status status;
@@ -974,16 +1009,38 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
 
   if (*cwd_state == NULL)
     {
-      *cwd_state = XMALLOC (struct saved_cwd, 1);
-      if (save_cwd (*cwd_state))
+      *cwd_state = xmalloc (sizeof **cwd_state);
+
+      if (save_cwd (&(*cwd_state)->saved_cwd) != 0)
 	{
-	  error (0, errno, _("cannot get current directory"));
-	  return RM_ERROR;
+	  (*cwd_state)->saved_errno = errno;
+	  assert (errno != 0);
+
+	  /* Pretend we started from "/".  That is fine as long as there
+	     is no requirement to return to the original working directory.
+	     Use "/", not ".", so that we chdir out of a non-root target
+	     directory before attempting to remove it: some hosts don't let
+	     you remove a working directory.  */
+	  (*cwd_state)->saved_cwd.name = xstrdup ("/");
 	}
-      AD_push_initial (ds, *cwd_state);
+      else
+	(*cwd_state)->saved_errno = 0;
+
+      AD_push_initial (ds);
       AD_INIT_OTHER_MEMBERS ();
     }
 
+  /* If we've failed to record and/or restore the initial working directory,
+     and we're now trying to access a `.'-relative file name, then give a
+     diagnostic, record the failure, and proceed with any subsequent
+     command-line arguments.  */
+  if ((*cwd_state)->saved_errno && IS_RELATIVE_FILE_NAME (dir))
+    {
+      error (0, (*cwd_state)->saved_errno, _("cannot remove directory %s"),
+	     quote (full_filename (dir)));
+      longjmp (ds->current_arg_jumpbuf, 1);
+    }
+
   /* There is a race condition in that an attacker could replace the nonempty
      directory, DIR, with a symlink between the preceding call to rmdir
      (in our caller) and the chdir below.  However, the following lstat,
@@ -997,14 +1054,11 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
       return RM_ERROR;
     }
 
-  if (chdir (dir))
+  errno = ENOTDIR;
+  if (! S_ISDIR (dir_sb.st_mode) || chdir (dir) != 0)
     {
-      if (! S_ISDIR (dir_sb.st_mode))
+      if (errno == ENOTDIR)
 	{
-	  /* This happens on Linux-2.4.18 when a non-privileged user tries
-	     to delete a file that is owned by another user in a directory
-	     like /tmp that has the S_ISVTX flag set.  */
-	  assert (saved_errno == EPERM);
 	  error (0, saved_errno,
 		 _("cannot remove %s"), quote (full_filename (dir)));
 	}
@@ -1017,6 +1071,12 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
       return RM_ERROR;
     }
 
+  if (ROOT_DEV_INO_CHECK (x->root_dev_ino, &dir_sb))
+    {
+      ROOT_DEV_INO_WARN (full_filename (dir));
+      return 1;
+    }
+
   AD_push (ds, dir, &dir_sb);
   AD_INIT_OTHER_MEMBERS ();
 
@@ -1045,7 +1105,11 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
       /* Execution reaches this point when we've removed the last
 	 removable entry from the current directory.  */
       {
-	char *d = AD_pop_and_chdir (ds);
+	/* This is the name of the directory that we have just
+	   returned from, after nominally removing all of its contents.  */
+	char *empty_dir;
+
+	AD_pop_and_chdir (ds, &empty_dir, *cwd_state);
 
 	/* Try to remove D only if remove_cwd_entries succeeded.  */
 	if (tmp_status == RM_OK)
@@ -1056,32 +1120,32 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
 	       But that's no big deal since we're interactive.  */
 	    Ternary is_dir;
 	    Ternary is_empty;
-	    enum RM_status s = prompt (ds, d, x, PA_REMOVE_DIR,
+	    enum RM_status s = prompt (ds, empty_dir, x, PA_REMOVE_DIR,
 				       &is_dir, &is_empty);
 
 	    if (s != RM_OK)
 	      {
-		free (d);
+		free (empty_dir);
 		return s;
 	      }
 
-	    if (rmdir (d) == 0)
+	    if (rmdir (empty_dir) == 0)
 	      {
 		if (x->verbose)
 		  printf (_("removed directory: %s\n"),
-			  quote (full_filename (d)));
+			  quote (full_filename (empty_dir)));
 	      }
 	    else
 	      {
 		error (0, errno, _("cannot remove directory %s"),
-		       quote (full_filename (d)));
-		AD_mark_as_unremovable (ds, d);
+		       quote (full_filename (empty_dir)));
+		AD_mark_as_unremovable (ds, empty_dir);
 		status = RM_ERROR;
 		UPDATE_STATUS (AD_stack_top(ds)->status, status);
 	      }
 	  }
 
-	free (d);
+	free (empty_dir);
 
 	if (AD_stack_height (ds) == 1)
 	  break;
@@ -1100,7 +1164,7 @@ remove_dir (Dirstack_state *ds, char const *dir, struct saved_cwd **cwd_state,
 
 static enum RM_status
 rm_1 (Dirstack_state *ds, char const *filename,
-      struct rm_options const *x, struct saved_cwd **cwd_state)
+      struct rm_options const *x, struct cwd_state **cwd_state)
 {
   char *base = base_name (filename);
   enum RM_status status;
@@ -1111,6 +1175,14 @@ rm_1 (Dirstack_state *ds, char const *filename,
       return RM_ERROR;
     }
 
+  if (*cwd_state && (*cwd_state)->saved_errno
+      && IS_RELATIVE_FILE_NAME (filename))
+    {
+      error (0, (*cwd_state)->saved_errno,
+	     _("cannot remove %s"), quote (filename));
+      return RM_ERROR;
+    }
+
   status = remove_entry (ds, filename, x, NULL);
   if (status != RM_NONEMPTY_DIR)
     return status;
@@ -1119,11 +1191,13 @@ rm_1 (Dirstack_state *ds, char const *filename,
 }
 
 /* Remove all files and/or directories specified by N_FILES and FILE.
-   Apply the options in X.  */
-enum RM_status
+   Apply the options in X.  If X->require_restore_cwd is false, then
+   this function may return RM_OK even though it is unable to restore
+   the initial working directory.  */
+extern enum RM_status
 rm (size_t n_files, char const *const *file, struct rm_options const *x)
 {
-  struct saved_cwd *cwd_state = NULL;
+  struct cwd_state *cwd_state = NULL;
   Dirstack_state *ds;
 
   /* Put the following two variables in static storage, so they can't
@@ -1148,9 +1222,19 @@ rm (size_t n_files, char const *const *file, struct rm_options const *x)
       UPDATE_STATUS (status, s);
     }
 
+  if (x->require_restore_cwd && cwd_state && cwd_state->saved_errno != 0)
+    {
+      error (0, cwd_state->saved_errno,
+	     _("cannot restore current working directory"));
+      status = RM_ERROR;
+    }
+
   ds_free (ds);
 
-  XFREE (cwd_state);
+  if (cwd_state && cwd_state->saved_errno == 0)
+    free_cwd (&cwd_state->saved_cwd);
+
+  free (cwd_state);
 
   return status;
 }
