@@ -19,7 +19,7 @@
    The author may be reached (Email) at the address mike@gnu.ai.mit.edu,
    or (US mail) as Mike Haertel c/o Free Software Foundation.
 
-   rn E. Hansen added NLS support in 1997.  */
+   Ørn E. Hansen added NLS support in 1997.  */
 
 #include <config.h>
 
@@ -31,9 +31,9 @@
 #include "error.h"
 #include "hard-locale.h"
 #include "inttostr.h"
-#include "long-options.h"
 #include "physmem.h"
 #include "posixver.h"
+#include "quote.h"
 #include "stdio-safer.h"
 #include "xmemcoll.h"
 #include "xstrtol.h"
@@ -65,7 +65,6 @@ double strtod ();
 #endif
 
 #define UCHAR_LIM (UCHAR_MAX + 1)
-#define UCHAR(c) ((unsigned char) (c))
 
 #ifndef DEFAULT_TMPDIR
 # define DEFAULT_TMPDIR "/tmp"
@@ -83,31 +82,22 @@ enum
     SORT_FAILURE = 2
   };
 
-#define C_DECIMAL_POINT '.'
 #define NEGATION_SIGN   '-'
 #define NUMERIC_ZERO    '0'
 
-#if HAVE_SETLOCALE
-
+/* The representation of the decimal point in the current locale.  */
 static char decimal_point;
-static int th_sep; /* if CHAR_MAX + 1, then there is no thousands separator */
+
+/* Thousands separator; if CHAR_MAX + 1, then there isn't one.  */
+static int thousands_sep;
 
 /* Nonzero if the corresponding locales are hard.  */
 static bool hard_LC_COLLATE;
-# if HAVE_NL_LANGINFO
+#if HAVE_NL_LANGINFO
 static bool hard_LC_TIME;
-# endif
-
-# define IS_THOUSANDS_SEP(x) ((x) == th_sep)
-
-#else
-
-# define decimal_point C_DECIMAL_POINT
-# define IS_THOUSANDS_SEP(x) 0
-
 #endif
 
-#define NONZERO(x) (x != 0)
+#define NONZERO(x) ((x) != 0)
 
 /* The kind of blanks for '-b' to skip in various options. */
 enum blanktype { bl_start, bl_end, bl_both };
@@ -148,8 +138,8 @@ struct keyfield
   size_t echar;			/* Additional characters in field. */
   bool const *ignore;		/* Boolean array of characters to ignore. */
   char const *translate;	/* Translation applied to characters. */
-  bool skipsblanks;		/* Skip leading blanks at start. */
-  bool skipeblanks;		/* Skip trailing blanks at finish. */
+  bool skipsblanks;		/* Skip leading blanks when finding start.  */
+  bool skipeblanks;		/* Skip leading blanks when finding end.  */
   bool numeric;			/* Flag for numeric comparison.  Handle
 				   strings of digits with optional decimal
 				   point, but no exponential notation. */
@@ -265,6 +255,17 @@ static struct keyfield *keylist;
 
 static void sortlines_temp (struct line *, size_t, struct line *);
 
+/* Report MESSAGE for FILE, then clean up and exit.
+   If FILE is null, it represents standard output.  */
+
+static void die (char const *, char const *) ATTRIBUTE_NORETURN;
+static void
+die (char const *message, char const *file)
+{
+  error (0, errno, "%s: %s", message, file ? file : _("standard output"));
+  exit (SORT_FAILURE);
+}
+
 void
 usage (int status)
 {
@@ -311,10 +312,10 @@ Other options:\n\
 "), stdout);
       printf (_("\
   -t, --field-separator=SEP use SEP instead of non-blank to blank transition\n\
-  -T, --temporary-directory=DIR  use DIR for temporaries, not $TMPDIR or %s\n\
+  -T, --temporary-directory=DIR  use DIR for temporaries, not $TMPDIR or %s;\n\
                               multiple options specify multiple directories\n\
-  -u, --unique              with -c: check for strict ordering\n\
-                              otherwise: output only the first of an equal run\n\
+  -u, --unique              with -c, check for strict ordering;\n\
+                              without -c, output only the first of an equal run\n\
 "), DEFAULT_TMPDIR);
       fputs (_("\
   -z, --zero-terminated     end lines with 0 byte, not newline\n\
@@ -383,8 +384,9 @@ struct tempnode
   char name[1];  /* Actual size is 1 + file name length.  */
 };
 static struct tempnode *volatile temphead;
+static struct tempnode *volatile *temptail = &temphead;
 
-/* Clean up any remaining temporary files. */
+/* Clean up any remaining temporary files.  */
 
 static void
 cleanup (void)
@@ -395,16 +397,6 @@ cleanup (void)
     unlink (node->name);
 }
 
-/* Report MESSAGE for FILE, then clean up and exit.  */
-
-static void die (char const *, char const *) ATTRIBUTE_NORETURN;
-static void
-die (char const *message, char const *file)
-{
-  error (0, errno, "%s: %s", message, file);
-  exit (SORT_FAILURE);
-}
-
 /* Create a new temporary file, returning its newly allocated name.
    Store into *PFP a stream open for writing.  */
 
@@ -419,12 +411,12 @@ create_temp_file (FILE **pfp)
   char const *temp_dir = temp_dirs[temp_dir_index];
   size_t len = strlen (temp_dir);
   struct tempnode *node =
-    xmalloc (sizeof node->next + len + sizeof slashbase);
+    xmalloc (offsetof (struct tempnode, name) + len + sizeof slashbase);
   char *file = node->name;
 
   memcpy (file, temp_dir, len);
   memcpy (file + len, slashbase, sizeof slashbase);
-  node->next = temphead;
+  node->next = NULL;
   if (++temp_dir_index == temp_dir_count)
     temp_dir_index = 0;
 
@@ -432,7 +424,10 @@ create_temp_file (FILE **pfp)
   sigprocmask (SIG_BLOCK, &caught_signals, &oldset);
   fd = mkstemp (file);
   if (0 <= fd)
-    temphead = node;
+    {
+      *temptail = node;
+      temptail = &node->next;
+    }
   saved_errno = errno;
   sigprocmask (SIG_SETMASK, &oldset, NULL);
   errno = saved_errno;
@@ -443,20 +438,22 @@ create_temp_file (FILE **pfp)
   return file;
 }
 
+/* Return a stream for FILE, opened with mode HOW.  A null FILE means
+   standard output; HOW should be "w".  When opening for input, "-"
+   means standard input.  To avoid confusion, do not return file
+   descriptors 0, 1, or 2.  */
+
 static FILE *
 xfopen (const char *file, const char *how)
 {
   FILE *fp;
 
-  if (STREQ (file, "-"))
+  if (!file)
+    fp = stdout;
+  else if (STREQ (file, "-") && *how == 'r')
     {
-      if (*how == 'r')
-	{
-	  have_read_stdin = true;
-	  fp = stdin;
-	}
-      else
-	fp = stdout;
+      have_read_stdin = true;
+      fp = stdin;
     }
   else
     {
@@ -478,6 +475,12 @@ xfclose (FILE *fp, char const *file)
       if (feof (fp))
 	clearerr (fp);
     }
+  else if (fp == stdout)
+    {
+      /* Don't close stdout just yet.  close_stdout does that.  */
+      if (fflush (fp) != 0)
+	die (_("fflush failed"), file);
+    }
   else
     {
       if (fclose (fp) != 0)
@@ -502,23 +505,34 @@ add_temp_dir (char const *dir)
   temp_dirs[temp_dir_count++] = dir;
 }
 
-/* Search through the list of temporary files for NAME;
-   remove it if it is found on the list. */
+/* Remove NAME from the list of temporary files.  */
 
 static void
 zaptemp (const char *name)
 {
   struct tempnode *volatile *pnode;
   struct tempnode *node;
+  struct tempnode *next;
+  sigset_t oldset;
+  int unlink_status;
+  int unlink_errno = 0;
 
-  for (pnode = &temphead; (node = *pnode); pnode = &node->next)
-    if (node->name == name)
-      {
-	unlink (name);
-	*pnode = node->next;
-	free (node);
-	break;
-      }
+  for (pnode = &temphead; (node = *pnode)->name != name; pnode = &node->next)
+    continue;
+
+  /* Unlink the temporary file in a critical section to avoid races.  */
+  next = node->next;
+  sigprocmask (SIG_BLOCK, &caught_signals, &oldset);
+  unlink_status = unlink (name);
+  unlink_errno = errno;
+  *pnode = next;
+  sigprocmask (SIG_SETMASK, &oldset, NULL);
+
+  if (unlink_status != 0)
+    error (0, unlink_errno, _("warning: cannot remove: %s"), name);
+  if (! next)
+    temptail = pnode;
+  free (node);
 }
 
 #if HAVE_NL_LANGINFO
@@ -538,7 +552,7 @@ struct_month_cmp (const void *m1, const void *m2)
 static void
 inittables (void)
 {
-  int i;
+  size_t i;
 
   for (i = 0; i < UCHAR_LIM; ++i)
     {
@@ -565,7 +579,7 @@ inittables (void)
 	  monthtab[i].val = i + 1;
 
 	  for (j = 0; j < s_len; j++)
-	    name[j] = fold_toupper[UCHAR (s[j])];
+	    name[j] = fold_toupper[to_uchar (s[j])];
 	  name[j] = '\0';
 	}
       qsort ((void *) monthtab, MONTHS_PER_YEAR,
@@ -684,8 +698,8 @@ default_sort_size (void)
    not specified by the user, use a default.  */
 
 static size_t
-sort_buffer_size (FILE *const *fps, int nfps,
-		  char *const *files, int nfiles,
+sort_buffer_size (FILE *const *fps, size_t nfps,
+		  char *const *files, size_t nfiles,
 		  size_t line_bytes)
 {
   /* A bound on the input size.  If zero, the bound hasn't been
@@ -699,7 +713,7 @@ sort_buffer_size (FILE *const *fps, int nfps,
      This extra room might be needed when preparing to read EOF.  */
   size_t size = worst_case_per_input_byte + 1;
 
-  int i;
+  size_t i;
 
   for (i = 0; i < nfiles; i++)
     {
@@ -708,7 +722,7 @@ sort_buffer_size (FILE *const *fps, int nfps,
       size_t worst_case;
 
       if ((i < nfps ? fstat (fileno (fps[i]), &st)
-	   : strcmp (files[i], "-") == 0 ? fstat (STDIN_FILENO, &st)
+	   : STREQ (files[i], "-") ? fstat (STDIN_FILENO, &st)
 	   : stat (files[i], &st))
 	  != 0)
 	die (_("stat failed"), files[i]);
@@ -805,14 +819,14 @@ begfield (const struct line *line, const struct keyfield *key)
   else
     while (ptr < lim && sword--)
       {
-	while (ptr < lim && blanks[UCHAR (*ptr)])
+	while (ptr < lim && blanks[to_uchar (*ptr)])
 	  ++ptr;
-	while (ptr < lim && !blanks[UCHAR (*ptr)])
+	while (ptr < lim && !blanks[to_uchar (*ptr)])
 	  ++ptr;
       }
 
   if (key->skipsblanks)
-    while (ptr < lim && blanks[UCHAR (*ptr)])
+    while (ptr < lim && blanks[to_uchar (*ptr)])
       ++ptr;
 
   /* Advance PTR by SCHAR (if possible), but no further than LIM.  */
@@ -853,9 +867,9 @@ limfield (const struct line *line, const struct keyfield *key)
   else
     while (ptr < lim && eword--)
       {
-	while (ptr < lim && blanks[UCHAR (*ptr)])
+	while (ptr < lim && blanks[to_uchar (*ptr)])
 	  ++ptr;
-	while (ptr < lim && !blanks[UCHAR (*ptr)])
+	while (ptr < lim && !blanks[to_uchar (*ptr)])
 	  ++ptr;
       }
 
@@ -902,18 +916,19 @@ limfield (const struct line *line, const struct keyfield *key)
     {
       char *newlim;
       newlim = ptr;
-      while (newlim < lim && blanks[UCHAR (*newlim)])
+      while (newlim < lim && blanks[to_uchar (*newlim)])
 	++newlim;
-      while (newlim < lim && !blanks[UCHAR (*newlim)])
+      while (newlim < lim && !blanks[to_uchar (*newlim)])
 	++newlim;
       lim = newlim;
     }
 #endif
 
-  /* If we're skipping leading blanks, don't start counting characters
-     until after skipping past any leading blanks.  */
-  if (key->skipsblanks)
-    while (ptr < lim && blanks[UCHAR (*ptr)])
+  /* If we're ignoring leading blanks when computing the End
+     of the field, don't start counting bytes until after skipping
+     past any leading blanks. */
+  if (key->skipeblanks)
+    while (ptr < lim && blanks[to_uchar (*ptr)])
       ++ptr;
 
   /* Advance PTR by ECHAR (if possible), but no further than LIM.  */
@@ -926,17 +941,6 @@ limfield (const struct line *line, const struct keyfield *key)
   return ptr;
 }
 
-/* Return the number of trailing blanks in FIELD, with LEN bytes.  */
-
-static size_t
-trailing_blanks (char const *field, size_t len)
-{
-  size_t i;
-  for (i = len; 0 < i && blanks[UCHAR (field[i - 1])]; i--)
-    continue;
-  return len - i;
-}
-
 /* Fill BUF reading from FP, moving buf->left bytes from the end
    of buf->buf to the beginning first.  If EOF is reached and the
    file wasn't terminated by a newline, supply one.  Set up BUF's line
@@ -1009,7 +1013,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
 	      if (key)
 		{
 		  /* Precompute the position of the first key for
-                     efficiency. */
+		     efficiency.  */
 		  line->keylim = (key->eword == SIZE_MAX
 				  ? p
 				  : limfield (line, key));
@@ -1019,15 +1023,10 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
 		  else
 		    {
 		      if (key->skipsblanks)
-			while (blanks[UCHAR (*line_start)])
+			while (blanks[to_uchar (*line_start)])
 			  line_start++;
 		      line->keybeg = line_start;
 		    }
-		  if (key->skipeblanks)
-		    {
-		      size_t keylen = line->keylim - line->keybeg;
-		      line->keylim -= trailing_blanks (line->keybeg, keylen);
-		    }
 		}
 
 	      line_start = ptr;
@@ -1125,22 +1124,22 @@ fraccompare (register const char *a, register const char *b)
 static int
 numcompare (register const char *a, register const char *b)
 {
-  register int tmpa, tmpb, tmp;
-  register size_t log_a, log_b;
+  char tmpa;
+  char tmpb;
+  int tmp;
+  size_t log_a;
+  size_t log_b;
 
-  tmpa = *a;
-  tmpb = *b;
-
-  while (blanks[UCHAR (tmpa)])
-    tmpa = *++a;
-  while (blanks[UCHAR (tmpb)])
-    tmpb = *++b;
+  while (blanks[to_uchar (tmpa = *a)])
+    a++;
+  while (blanks[to_uchar (tmpb = *b)])
+    b++;
 
   if (tmpa == NEGATION_SIGN)
     {
       do
 	tmpa = *++a;
-      while (tmpa == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpa));
+      while (tmpa == NUMERIC_ZERO || tmpa == thousands_sep);
       if (tmpb != NEGATION_SIGN)
 	{
 	  if (tmpa == decimal_point)
@@ -1149,45 +1148,43 @@ numcompare (register const char *a, register const char *b)
 	    while (tmpa == NUMERIC_ZERO);
 	  if (ISDIGIT (tmpa))
 	    return -1;
-	  while (tmpb == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpb))
+	  while (tmpb == NUMERIC_ZERO || tmpb == thousands_sep)
 	    tmpb = *++b;
 	  if (tmpb == decimal_point)
 	    do
 	      tmpb = *++b;
 	    while (tmpb == NUMERIC_ZERO);
-	  if (ISDIGIT (tmpb))
-	    return -1;
-	  return 0;
+	  return - ISDIGIT (tmpb);
 	}
       do
 	tmpb = *++b;
-      while (tmpb == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpb));
+      while (tmpb == NUMERIC_ZERO || tmpb == thousands_sep);
 
       while (tmpa == tmpb && ISDIGIT (tmpa))
 	{
 	  do
 	    tmpa = *++a;
-	  while (IS_THOUSANDS_SEP (tmpa));
+	  while (tmpa == thousands_sep);
 	  do
 	    tmpb = *++b;
-	  while (IS_THOUSANDS_SEP (tmpb));
+	  while (tmpb == thousands_sep);
 	}
 
       if ((tmpa == decimal_point && !ISDIGIT (tmpb))
 	  || (tmpb == decimal_point && !ISDIGIT (tmpa)))
-	return -fraccompare (a, b);
+	return fraccompare (b, a);
 
       tmp = tmpb - tmpa;
 
       for (log_a = 0; ISDIGIT (tmpa); ++log_a)
 	do
 	  tmpa = *++a;
-	while (IS_THOUSANDS_SEP (tmpa));
+	while (tmpa == thousands_sep);
 
       for (log_b = 0; ISDIGIT (tmpb); ++log_b)
 	do
 	  tmpb = *++b;
-	while (IS_THOUSANDS_SEP (tmpb));
+	while (tmpb == thousands_sep);
 
       if (log_a != log_b)
 	return log_a < log_b ? 1 : -1;
@@ -1201,38 +1198,36 @@ numcompare (register const char *a, register const char *b)
     {
       do
 	tmpb = *++b;
-      while (tmpb == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpb));
+      while (tmpb == NUMERIC_ZERO || tmpb == thousands_sep);
       if (tmpb == decimal_point)
 	do
 	  tmpb = *++b;
 	while (tmpb == NUMERIC_ZERO);
       if (ISDIGIT (tmpb))
 	return 1;
-      while (tmpa == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpa))
+      while (tmpa == NUMERIC_ZERO || tmpa == thousands_sep)
 	tmpa = *++a;
       if (tmpa == decimal_point)
 	do
 	  tmpa = *++a;
 	while (tmpa == NUMERIC_ZERO);
-      if (ISDIGIT (tmpa))
-	return 1;
-      return 0;
+      return ISDIGIT (tmpa);
     }
   else
     {
-      while (tmpa == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpa))
+      while (tmpa == NUMERIC_ZERO || tmpa == thousands_sep)
 	tmpa = *++a;
-      while (tmpb == NUMERIC_ZERO || IS_THOUSANDS_SEP (tmpb))
+      while (tmpb == NUMERIC_ZERO || tmpb == thousands_sep)
 	tmpb = *++b;
 
       while (tmpa == tmpb && ISDIGIT (tmpa))
 	{
 	  do
 	    tmpa = *++a;
-	  while (IS_THOUSANDS_SEP (tmpa));
+	  while (tmpa == thousands_sep);
 	  do
 	    tmpb = *++b;
-	  while (IS_THOUSANDS_SEP (tmpb));
+	  while (tmpb == thousands_sep);
 	}
 
       if ((tmpa == decimal_point && !ISDIGIT (tmpb))
@@ -1244,12 +1239,12 @@ numcompare (register const char *a, register const char *b)
       for (log_a = 0; ISDIGIT (tmpa); ++log_a)
 	do
 	  tmpa = *++a;
-	while (IS_THOUSANDS_SEP (tmpa));
+	while (tmpa == thousands_sep);
 
       for (log_b = 0; ISDIGIT (tmpb); ++log_b)
 	do
 	  tmpb = *++b;
-	while (IS_THOUSANDS_SEP (tmpb));
+	while (tmpb == thousands_sep);
 
       if (log_a != log_b)
 	return log_a < log_b ? -1 : 1;
@@ -1290,46 +1285,50 @@ general_numcompare (const char *sa, const char *sb)
 	  : memcmp ((char *) &a, (char *) &b, sizeof a));
 }
 
-/* Return an integer in 1..12 of the month name S with length LEN.
+/* Return an integer in 1..12 of the month name MONTH with length LEN.
    Return 0 if the name in S is not recognized.  */
 
 static int
-getmonth (const char *s, size_t len)
+getmonth (char const *month, size_t len)
 {
-  char *month;
-  register size_t i;
-  register int lo = 0, hi = MONTHS_PER_YEAR, result;
+  size_t lo = 0;
+  size_t hi = MONTHS_PER_YEAR;
+  char const *monthlim = month + len;
 
-  while (len > 0 && blanks[UCHAR (*s)])
+  for (;;)
     {
-      ++s;
-      --len;
+      if (month == monthlim)
+	return 0;
+      if (!blanks[to_uchar (*month)])
+	break;
+      ++month;
     }
 
-  if (len == 0)
-    return 0;
-
-  month = alloca (len + 1);
-  for (i = 0; i < len; ++i)
-    month[i] = fold_toupper[UCHAR (s[i])];
-  len -= trailing_blanks (month, len);
-  month[len] = '\0';
-
   do
     {
-      int ix = (lo + hi) / 2;
+      size_t ix = (lo + hi) / 2;
+      char const *m = month;
+      char const *n = monthtab[ix].name;
 
-      if (strncmp (month, monthtab[ix].name, strlen (monthtab[ix].name)) < 0)
-	hi = ix;
-      else
-	lo = ix;
+      for (;; m++, n++)
+	{
+	  if (!*n)
+	    return monthtab[ix].val;
+	  if (m == monthlim || fold_toupper[to_uchar (*m)] < to_uchar (*n))
+	    {
+	      hi = ix;
+	      break;
+	    }
+	  else if (fold_toupper[to_uchar (*m)] > to_uchar (*n))
+	    {
+	      lo = ix + 1;
+	      break;
+	    }
+	}
     }
-  while (hi - lo > 1);
+  while (lo < hi);
 
-  result = (!strncmp (month, monthtab[lo].name, strlen (monthtab[lo].name))
-	    ? monthtab[lo].val : 0);
-
-  return result;
+  return 0;
 }
 
 /* Compare two lines A and B trying every key in sequence until there
@@ -1358,12 +1357,6 @@ keycompare (const struct line *a, const struct line *b)
       size_t lena = lima <= texta ? 0 : lima - texta;
       size_t lenb = limb <= textb ? 0 : limb - textb;
 
-      if (key->skipeblanks)
-	{
-	  lena -= trailing_blanks (texta, lena);
-	  lenb -= trailing_blanks (textb, lenb);
-	}
-
       /* Actually compare the fields. */
       if (key->numeric | key->general_numeric)
 	{
@@ -1377,12 +1370,14 @@ keycompare (const struct line *a, const struct line *b)
       else if (key->month)
 	diff = getmonth (texta, lena) - getmonth (textb, lenb);
       /* Sorting like this may become slow, so in a simple locale the user
-         can select a faster sort that is similar to ascii sort  */
-      else if (HAVE_SETLOCALE && hard_LC_COLLATE)
+	 can select a faster sort that is similar to ascii sort.  */
+      else if (hard_LC_COLLATE)
 	{
 	  if (ignore || translate)
 	    {
-	      char *copy_a = alloca (lena + 1 + lenb + 1);
+	      char buf[4000];
+	      size_t size = lena + 1 + lenb + 1;
+	      char *copy_a = (size <= sizeof buf ? buf : xmalloc (size));
 	      char *copy_b = copy_a + lena + 1;
 	      size_t new_len_a, new_len_b, i;
 
@@ -1392,22 +1387,25 @@ keycompare (const struct line *a, const struct line *b)
 		  if (i < lena)
 		    {
 		      copy_a[new_len_a] = (translate
-					   ? translate[UCHAR (texta[i])]
+					   ? translate[to_uchar (texta[i])]
 					   : texta[i]);
-		      if (!ignore || !ignore[UCHAR (texta[i])])
+		      if (!ignore || !ignore[to_uchar (texta[i])])
 			++new_len_a;
 		    }
 		  if (i < lenb)
 		    {
 		      copy_b[new_len_b] = (translate
-					   ? translate[UCHAR (textb[i])]
+					   ? translate[to_uchar (textb[i])]
 					   : textb [i]);
-		      if (!ignore || !ignore[UCHAR (textb[i])])
+		      if (!ignore || !ignore[to_uchar (textb[i])])
 			++new_len_b;
 		    }
 		}
 
 	      diff = xmemcoll (copy_a, new_len_a, copy_b, new_len_b);
+
+	      if (sizeof buf < size)
+		free (copy_a);
 	    }
 	  else if (lena == 0)
 	    diff = - NONZERO (lenb);
@@ -1423,13 +1421,13 @@ keycompare (const struct line *a, const struct line *b)
     {									\
 	  for (;;)							\
 	    {								\
-	      while (texta < lima && ignore[UCHAR (*texta)])		\
+	      while (texta < lima && ignore[to_uchar (*texta)])		\
 		++texta;						\
-	      while (textb < limb && ignore[UCHAR (*textb)])		\
+	      while (textb < limb && ignore[to_uchar (*textb)])		\
 		++textb;						\
 	      if (! (texta < lima && textb < limb))			\
 		break;							\
-	      diff = UCHAR (A) - UCHAR (B);				\
+	      diff = to_uchar (A) - to_uchar (B);			\
 	      if (diff)							\
 		goto not_equal;						\
 	      ++texta;							\
@@ -1441,8 +1439,8 @@ keycompare (const struct line *a, const struct line *b)
   while (0)
 
 	  if (translate)
-	    CMP_WITH_IGNORE (translate[UCHAR (*texta)],
-			     translate[UCHAR (*textb)]);
+	    CMP_WITH_IGNORE (translate[to_uchar (*texta)],
+			     translate[to_uchar (*textb)]);
 	  else
 	    CMP_WITH_IGNORE (*texta, *textb);
 	}
@@ -1456,8 +1454,8 @@ keycompare (const struct line *a, const struct line *b)
 	    {
 	      while (texta < lima && textb < limb)
 		{
-		  diff = (UCHAR (translate[UCHAR (*texta++)])
-			  - UCHAR (translate[UCHAR (*textb++)]));
+		  diff = (to_uchar (translate[to_uchar (*texta++)])
+			  - to_uchar (translate[to_uchar (*textb++)]));
 		  if (diff)
 		    goto not_equal;
 		}
@@ -1491,9 +1489,9 @@ keycompare (const struct line *a, const struct line *b)
 	  texta = a->text, textb = b->text;
 	  if (key->skipsblanks)
 	    {
-	      while (texta < lima && blanks[UCHAR (*texta)])
+	      while (texta < lima && blanks[to_uchar (*texta)])
 		++texta;
-	      while (textb < limb && blanks[UCHAR (*textb)])
+	      while (textb < limb && blanks[to_uchar (*textb)])
 		++textb;
 	    }
 	}
@@ -1522,7 +1520,6 @@ compare (register const struct line *a, register const struct line *b)
   if (keylist)
     {
       diff = keycompare (a, b);
-      alloca (0);
       if (diff | unique | stable)
 	return diff;
     }
@@ -1535,7 +1532,7 @@ compare (register const struct line *a, register const struct line *b)
     diff = - NONZERO (blen);
   else if (blen == 0)
     diff = 1;
-  else if (HAVE_SETLOCALE && hard_LC_COLLATE)
+  else if (hard_LC_COLLATE)
     diff = xmemcoll (a->text, alen, b->text, blen);
   else if (! (diff = memcmp (a->text, b->text, MIN (alen, blen))))
     diff = alen < blen ? -1 : alen != blen;
@@ -1628,14 +1625,17 @@ check (char const *file_name)
   return ordered;
 }
 
-/* Merge lines from FILES onto OFP.  NFILES cannot be greater than
-   NMERGE.  Close input and output files before returning.
-   OUTPUT_FILE gives the name of the output file; if OFP is NULL, the
-   output file has not been opened yet.  */
+/* Merge lines from FILES onto OFP.  NTEMPS is the number of temporary
+   files (all of which are at the start of the FILES array), and
+   NFILES is the number of files; 0 <= NTEMPS <= NFILES <= NMERGE.
+   Close input and output files before returning.
+   OUTPUT_FILE gives the name of the output file.  If it is NULL,
+   the output file is standard output.  If OFP is NULL, the output
+   file has not been opened yet (or written to, if standard output).  */
 
 static void
-mergefps (char **files, register int nfiles,
-	  FILE *ofp, const char *output_file)
+mergefps (char **files, size_t ntemps, size_t nfiles,
+	  FILE *ofp, char const *output_file)
 {
   FILE *fps[NMERGE];		/* Input streams for each file.  */
   struct buffer buffer[NMERGE];	/* Input buffers for each file. */
@@ -1645,10 +1645,12 @@ mergefps (char **files, register int nfiles,
   size_t savealloc = 0;		/* Size allocated for the saved line. */
   struct line const *cur[NMERGE]; /* Current line in each line table. */
   struct line const *base[NMERGE]; /* Base of each line table.  */
-  int ord[NMERGE];		/* Table representing a permutation of fps,
+  size_t ord[NMERGE];		/* Table representing a permutation of fps,
 				   such that cur[ord[0]] is the smallest line
 				   and will be next output. */
-  register int i, j, t;
+  size_t i;
+  size_t j;
+  size_t t;
   struct keyfield const *key = keylist;
   saved.text = NULL;
 
@@ -1669,7 +1671,11 @@ mergefps (char **files, register int nfiles,
 	{
 	  /* fps[i] is empty; eliminate it from future consideration.  */
 	  xfclose (fps[i], files[i]);
-	  zaptemp (files[i]);
+	  if (i < ntemps)
+	    {
+	      ntemps--;
+	      zaptemp (files[i]);
+	    }
 	  free (buffer[i].buf);
 	  --nfiles;
 	  for (j = i; j < nfiles; ++j)
@@ -1745,13 +1751,17 @@ mergefps (char **files, register int nfiles,
 	    }
 	  else
 	    {
-	      /* We reached EOF on fps[ord[0]]. */
+	      /* We reached EOF on fps[ord[0]].  */
 	      for (i = 1; i < nfiles; ++i)
 		if (ord[i] > ord[0])
 		  --ord[i];
 	      --nfiles;
 	      xfclose (fps[ord[0]], files[ord[0]]);
-	      zaptemp (files[ord[0]]);
+	      if (ord[0] < ntemps)
+		{
+		  ntemps--;
+		  zaptemp (files[ord[0]]);
+		}
 	      free (buffer[ord[0]].buf);
 	      for (i = ord[0]; i < nfiles; ++i)
 		{
@@ -1772,10 +1782,8 @@ mergefps (char **files, register int nfiles,
 	 a line larger than it. */
       for (i = 1; i < nfiles; ++i)
 	{
-	  t = compare (cur[ord[0]], cur[ord[i]]);
-	  if (!t)
-	    t = ord[0] - ord[i];
-	  if (t < 0)
+	  int cmp = compare (cur[ord[0]], cur[ord[i]]);
+	  if (cmp < 0 || (cmp == 0 && ord[0] < ord[i]))
 	    break;
 	}
       t = ord[0];
@@ -1899,82 +1907,141 @@ sortlines_temp (struct line *lines, size_t nlines, struct line *temp)
     }
 }
 
-/* Return the index of the first of NFILES FILES that is the same file
-   as OUTFILE.  If none can be the same, return NFILES.  Consider an
-   input pipe to be the same as OUTFILE, since the pipe might be the
-   output of a command like "cat OUTFILE".  */
+/* Scan through FILES[NTEMPS .. NFILES-1] looking for a file that is
+   the same as OUTFILE.  If found, merge the found instances (and perhaps
+   some other files) into a temporary file so that it can in turn be
+   merged into OUTFILE without destroying OUTFILE before it is completely
+   read.  Return the new value of NFILES, which differs from the old if
+   some merging occurred.
 
-static int
-first_same_file (char * const *files, int nfiles, char const *outfile)
+   This test ensures that an otherwise-erroneous use like
+   "sort -m -o FILE ... FILE ..." copies FILE before writing to it.
+   It's not clear that POSIX requires this nicety.
+   Detect common error cases, but don't try to catch obscure cases like
+   "cat ... FILE ... | sort -m -o FILE"
+   where traditional "sort" doesn't copy the input and where
+   people should know that they're getting into trouble anyway.
+   Catching these obscure cases would slow down performance in
+   common cases.  */
+
+static size_t
+avoid_trashing_input (char **files, size_t ntemps, size_t nfiles,
+		      char const *outfile)
 {
-  int i;
+  size_t i;
   bool got_outstat = false;
-  struct stat instat, outstat;
+  struct stat outstat;
 
-  for (i = 0; i < nfiles; i++)
+  for (i = ntemps; i < nfiles; i++)
     {
       bool standard_input = STREQ (files[i], "-");
+      bool same;
+      struct stat instat;
 
-      if (STREQ (outfile, files[i]) && ! standard_input)
-	return i;
-
-      if (! got_outstat)
+      if (outfile && STREQ (outfile, files[i]) && ! standard_input)
+	same = true;
+      else
 	{
-	  got_outstat = true;
-	  if ((STREQ (outfile, "-")
-	       ? fstat (STDOUT_FILENO, &outstat)
-	       : stat (outfile, &outstat))
-	      != 0)
-	    return nfiles;
+	  if (! got_outstat)
+	    {
+	      if ((outfile
+		   ? stat (outfile, &outstat)
+		   : fstat (STDOUT_FILENO, &outstat))
+		  != 0)
+		break;
+	      got_outstat = true;
+	    }
+
+	  same = (((standard_input
+		    ? fstat (STDIN_FILENO, &instat)
+		    : stat (files[i], &instat))
+		   == 0)
+		  && SAME_INODE (instat, outstat));
 	}
 
-      if (((standard_input
-	    ? fstat (STDIN_FILENO, &instat)
-	    : stat (files[i], &instat))
-	   == 0)
-	  && (S_ISFIFO (instat.st_mode) || SAME_INODE (instat, outstat)))
-	return i;
+      if (same)
+	{
+	  FILE *tftp;
+	  char *temp = create_temp_file (&tftp);
+	  mergefps (&files[i], 0, nfiles - i, tftp, temp);
+	  files[i] = temp;
+	  return i + 1;
+	}
     }
 
   return nfiles;
 }
 
-/* Merge NFILES FILES onto OUTPUT_FILE.  However, merge at most
-   MAX_MERGE input files directly onto OUTPUT_FILE.  MAX_MERGE cannot
-   exceed NMERGE.  */
+/* Merge the input FILES.  NTEMPS is the number of files at the
+   start of FILES that are temporary; it is zero at the top level.
+   NFILES is the total number of files.  Put the output in
+   OUTPUT_FILE; a null OUTPUT_FILE stands for standard output.  */
 
 static void
-merge (char **files, int nfiles, int max_merge, char const *output_file)
+merge (char **files, size_t ntemps, size_t nfiles, char const *output_file)
 {
-  while (max_merge < nfiles)
+  while (NMERGE < nfiles)
     {
-      FILE *tfp;
-      int i, t = 0;
-      char *temp;
-      for (i = 0; i < nfiles / NMERGE; ++i)
+      /* Number of input files processed so far.  */
+      size_t in;
+
+      /* Number of output files generated so far.  */
+      size_t out;
+
+      /* nfiles % NMERGE; this counts input files that are left over
+	 after all full-sized merges have been done.  */
+      size_t remainder;
+
+      /* Number of easily-available slots at the next loop iteration.  */
+      size_t cheap_slots;
+
+      /* Do as many NMERGE-size merges as possible.  */
+      for (out = in = 0; out < nfiles / NMERGE; out++, in += NMERGE)
 	{
-	  temp = create_temp_file (&tfp);
-	  mergefps (&files[i * NMERGE], NMERGE, tfp, temp);
-	  files[t++] = temp;
+	  FILE *tfp;
+	  char *temp = create_temp_file (&tfp);
+	  size_t nt = MIN (ntemps, NMERGE);
+	  ntemps -= nt;
+	  mergefps (&files[in], nt, NMERGE, tfp, temp);
+	  files[out] = temp;
 	}
-      temp = create_temp_file (&tfp);
-      mergefps (&files[i * NMERGE], nfiles % NMERGE, tfp, temp);
-      files[t++] = temp;
-      nfiles = t;
-      if (nfiles == 1)
-	break;
+
+      remainder = nfiles - in;
+      cheap_slots = NMERGE - out % NMERGE;
+
+      if (cheap_slots < remainder)
+	{
+	  /* So many files remain that they can't all be put into the last
+	     NMERGE-sized output window.  Do one more merge.  Merge as few
+	     files as possible, to avoid needless I/O.  */
+	  size_t nshortmerge = remainder - cheap_slots + 1;
+	  FILE *tfp;
+	  char *temp = create_temp_file (&tfp);
+	  size_t nt = MIN (ntemps, nshortmerge);
+	  ntemps -= nt;
+	  mergefps (&files[in], nt, nshortmerge, tfp, temp);
+	  files[out++] = temp;
+	  in += nshortmerge;
+	}
+
+      /* Put the remaining input files into the last NMERGE-sized output
+	 window, so they will be merged in the next pass.  */
+      memmove(&files[out], &files[in], (nfiles - in) * sizeof *files);
+      ntemps += out;
+      nfiles -= in - out;
     }
 
-  mergefps (files, nfiles, NULL, output_file);
+  nfiles = avoid_trashing_input (files, ntemps, nfiles, output_file);
+  mergefps (files, ntemps, nfiles, NULL, output_file);
 }
 
 /* Sort NFILES FILES onto OUTPUT_FILE. */
 
 static void
-sort (char * const *files, int nfiles, char const *output_file)
+sort (char * const *files, size_t nfiles, char const *output_file)
 {
   struct buffer buf;
-  int n_temp_files = 0;
+  size_t ntemps = 0;
   bool output_file_created = false;
 
   buf.alloc = 0;
@@ -2015,7 +2082,7 @@ sort (char * const *files, int nfiles, char const *output_file)
 	  linebase = line - buf.nlines;
 	  if (1 < buf.nlines)
 	    sortlines (line, buf.nlines, linebase);
-	  if (buf.eof && !nfiles && !n_temp_files && !buf.left)
+	  if (buf.eof && !nfiles && !ntemps && !buf.left)
 	    {
 	      xfclose (fp, file);
 	      tfp = xfopen (output_file, "w");
@@ -2024,7 +2091,7 @@ sort (char * const *files, int nfiles, char const *output_file)
 	    }
 	  else
 	    {
-	      ++n_temp_files;
+	      ++ntemps;
 	      temp_output = create_temp_file (&tfp);
 	    }
 
@@ -2051,12 +2118,15 @@ sort (char * const *files, int nfiles, char const *output_file)
 
   if (! output_file_created)
     {
-      int i = n_temp_files;
-      struct tempnode *node;
-      char **tempfiles = xnmalloc (n_temp_files, sizeof *tempfiles);
-      for (node = temphead; i > 0; node = node->next)
-	tempfiles[--i] = node->name;
-      merge (tempfiles, n_temp_files, NMERGE, output_file);
+      size_t i;
+      struct tempnode *node = temphead;
+      char **tempfiles = xnmalloc (ntemps, sizeof *tempfiles);
+      for (i = 0; node; i++)
+	{
+	  tempfiles[i] = node->name;
+	  node = node->next;
+	}
+      merge (tempfiles, ntemps, ntemps, output_file);
       free (tempfiles);
     }
 }
@@ -2133,19 +2203,7 @@ sighandler (int sig)
 
   cleanup ();
 
-#ifdef SA_NOCLDSTOP
-  {
-    struct sigaction sigact;
-
-    sigact.sa_handler = SIG_DFL;
-    sigemptyset (&sigact.sa_mask);
-    sigact.sa_flags = 0;
-    sigaction (sig, &sigact, NULL);
-  }
-#else
   signal (sig, SIG_DFL);
-#endif
-
   raise (sig);
 }
 
@@ -2217,19 +2275,14 @@ main (int argc, char **argv)
   int c = 0;
   bool checkonly = false;
   bool mergeonly = false;
-  int nfiles = 0;
+  size_t nfiles = 0;
   bool posixly_correct = (getenv ("POSIXLY_CORRECT") != NULL);
   bool obsolete_usage = (posix2_version () < 200112);
   char const *short_options = (obsolete_usage
 			       ? COMMON_SHORT_OPTIONS "y::"
 			       : COMMON_SHORT_OPTIONS "y:");
   char *minus = "-", **files;
-  char const *outfile = minus;
-  static int const sigs[] = { SIGHUP, SIGINT, SIGPIPE, SIGTERM };
-  unsigned int nsigs = sizeof sigs / sizeof *sigs;
-#ifdef SA_NOCLDSTOP
-  struct sigaction oldact, newact;
-#endif
+  char const *outfile = NULL;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -2247,54 +2300,54 @@ main (int argc, char **argv)
   hard_LC_TIME = hard_locale (LC_TIME);
 #endif
 
-#if HAVE_SETLOCALE
-  /* Let's get locale's representation of the decimal point */
+  /* Get locale's representation of the decimal point.  */
   {
-    struct lconv const *lconvp = localeconv ();
+    struct lconv const *locale = localeconv ();
 
     /* If the locale doesn't define a decimal point, or if the decimal
-       point is multibyte, use the C decimal point.  We don't support
-       multibyte decimal points yet.  */
-    decimal_point = *lconvp->decimal_point;
-    if (! decimal_point || lconvp->decimal_point[1])
-      decimal_point = C_DECIMAL_POINT;
+       point is multibyte, use the C locale's decimal point.  FIXME:
+       add support for multibyte decimal points.  */
+    decimal_point = locale->decimal_point[0];
+    if (! decimal_point || locale->decimal_point[1])
+      decimal_point = '.';
 
-    /* We don't support multibyte thousands separators yet.  */
-    th_sep = *lconvp->thousands_sep;
-    if (! th_sep || lconvp->thousands_sep[1])
-      th_sep = CHAR_MAX + 1;
+    /* FIXME: add support for multibyte thousands separators.  */
+    thousands_sep = *locale->thousands_sep;
+    if (! thousands_sep || locale->thousands_sep[1])
+      thousands_sep = CHAR_MAX + 1;
   }
-#endif
 
   have_read_stdin = false;
   inittables ();
 
-#ifdef SA_NOCLDSTOP
   {
-    unsigned int i;
+    size_t i;
+    static int const sig[] = { SIGHUP, SIGINT, SIGPIPE, SIGTERM };
+    enum { nsigs = sizeof sig / sizeof sig[0] };
+
+#ifdef SA_NOCLDSTOP
+    struct sigaction act;
+
     sigemptyset (&caught_signals);
-    for (i = 0; i < nsigs; i++)
-      sigaddset (&caught_signals, sigs[i]);
-    newact.sa_handler = sighandler;
-    newact.sa_mask = caught_signals;
-    newact.sa_flags = 0;
-  }
-#endif
-
-  {
-    unsigned int i;
     for (i = 0; i < nsigs; i++)
       {
-	int sig = sigs[i];
-#ifdef SA_NOCLDSTOP
-	sigaction (sig, NULL, &oldact);
-	if (oldact.sa_handler != SIG_IGN)
-	  sigaction (sig, &newact, NULL);
-#else
-	if (signal (sig, SIG_IGN) != SIG_IGN)
-	  signal (sig, sighandler);
-#endif
+	sigaction (sig[i], NULL, &act);
+	if (act.sa_handler != SIG_IGN)
+	  sigaddset (&caught_signals, sig[i]);
       }
+
+    act.sa_handler = sighandler;
+    act.sa_mask = caught_signals;
+    act.sa_flags = 0;
+
+    for (i = 0; i < nsigs; i++)
+      if (sigismember (&caught_signals, sig[i]))
+	sigaction (sig[i], &act, NULL);
+#else
+    for (i = 0; i < nsigs; i++)
+      if (signal (sig[i], SIG_IGN) != SIG_IGN)
+	signal (sig[i], sighandler);
+#endif
   }
 
   gkey.sword = gkey.eword = SIZE_MAX;
@@ -2308,9 +2361,9 @@ main (int argc, char **argv)
   for (;;)
     {
       /* Parse an operand as a file after "--" was seen; or if
-         pedantic and a file was seen, unless the POSIX version
-         predates 1003.1-2001 and -c was not seen and the operand is
-         "-o FILE" or "-oFILE".  */
+	 pedantic and a file was seen, unless the POSIX version
+	 predates 1003.1-2001 and -c was not seen and the operand is
+	 "-o FILE" or "-oFILE".  */
 
       if (c == -1
 	  || (posixly_correct && nfiles != 0
@@ -2447,7 +2500,7 @@ main (int argc, char **argv)
 	  break;
 
 	case 'o':
-	  if (outfile != minus && strcmp (outfile, optarg) != 0)
+	  if (outfile && !STREQ (outfile, optarg))
 	    error (SORT_FAILURE, 0, _("multiple output files specified"));
 	  outfile = optarg;
 	  break;
@@ -2462,12 +2515,12 @@ main (int argc, char **argv)
 
 	case 't':
 	  {
-	    int newtab = optarg[0];
+	    char newtab = optarg[0];
 	    if (! newtab)
 	      error (SORT_FAILURE, 0, _("empty tab"));
 	    if (optarg[1])
 	      {
-		if (strcmp (optarg, "\\0") == 0)
+		if (STREQ (optarg, "\\0"))
 		  newtab = '\0';
 		else
 		  {
@@ -2494,9 +2547,23 @@ main (int argc, char **argv)
 	  break;
 
 	case 'y':
-	  /* Accept and ignore e.g. -y0 for compatibility with Solaris
-	     2.x through Solaris 7.  -y is marked as obsolete starting
-	     with Solaris 8.  */
+	  /* Accept and ignore e.g. -y0 for compatibility with Solaris 2.x
+	     through Solaris 7.  It is also accepted by many non-Solaris
+	     "sort" implementations, e.g., AIX 5.2, HP-UX 11i v2, IRIX 6.5.
+	     -y is marked as obsolete starting with Solaris 8 (1999), but is
+	     still accepted as of Solaris 10 prerelease (2004).
+
+	     Solaris 2.5.1 "sort -y 100" reads the input file "100", but
+	     emulate Solaris 8 and 9 "sort -y 100" which ignores the "100",
+	     and which in general ignores the argument after "-y" if it
+	     consists entirely of digits (it can even be empty).  */
+	  if (!optarg && optind != argc)
+	    {
+	      char const *p;
+	      for (p = argv[optind]; ISDIGIT (*p); p++)
+		continue;
+	      optind += !*p;
+	    }
 	  break;
 
 	case 'z':
@@ -2550,8 +2617,11 @@ main (int argc, char **argv)
   if (checkonly)
     {
       if (nfiles > 1)
-	error (SORT_FAILURE, 0, _("extra operand `%s' not allowed with -c"),
-	       files[1]);
+	{
+	  error (0, 0, _("extra operand %s not allowed with -c"),
+		 quote (files[1]));
+	  usage (SORT_FAILURE);
+	}
 
       /* POSIX requires that sort return 1 IFF invoked with -c and the
 	 input is not properly sorted.  */
@@ -2559,10 +2629,7 @@ main (int argc, char **argv)
     }
 
   if (mergeonly)
-    {
-      int max_merge = first_same_file (files, MIN (nfiles, NMERGE), outfile);
-      merge (files, nfiles, max_merge, outfile);
-    }
+    merge (files, 0, nfiles, outfile);
   else
     sort (files, nfiles, outfile);
 
